# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import argparse
import entropy
import typing

from decimal import Decimal

from .pairwiseelement import PairwiseElement
from ...modelstate import ModelState


# # 🥭 BiasQuantityOnPositionElement class
#
# Modifies an `Order`s quantity based on current inventory. Looks at the current position, the maximum desired
# position, and the target position. Tries to bias the order quantity to tend towards the target position,
# incresing the order quantities that shift it in that direction and decreasing the order quantities that
# shift it away from the target.
#
# For example(1):
#  BUY quantity: 10
#  SELL quantity: 10
#  CURRENT position: 0
#  MAXIMUM position: 50
#  TARGET position: 0
#  Result: no change
#
# For example(2):
#  BUY quantity: 10
#  SELL quantity: 10
#  CURRENT position: 20
#  MAXIMUM position: 50
#  TARGET position: 0
#  Result:
#    BUY quantity: 6
#    SELL quantity: 14
#
# In example (2) you can see:
# * the current position of 20 means BUY quantity has decreased and SELL quantity has increased
# * the total quantity in BUY and SELL stays the same, at 20
#
# More examples (including all data used in unit tests) are available in the
# docs/BiasQuantityOnPosition.ods spreadsheet.
#
class BiasQuantityOnPositionElement(PairwiseElement):
    def __init__(
        self, maximum_position: Decimal, target_position: Decimal = Decimal(0)
    ) -> None:
        super().__init__()
        self.maximum_position: Decimal = maximum_position
        self.target_position: Decimal = target_position

    @staticmethod
    def add_command_line_parameters(parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--biasquantityonposition-maximum-position",
            type=Decimal,
            help="maximum inventory position to proportionally move away from",
        )
        parser.add_argument(
            "--biasquantityonposition-target-position",
            type=Decimal,
            help="inventory position to target (default 0)",
        )

    @staticmethod
    def from_command_line_parameters(
        args: argparse.Namespace,
    ) -> "BiasQuantityOnPositionElement":
        if args.biasquantityonposition_maximum_position is None:
            raise Exception(
                "No maximum position value specified for biasing. Try the --biasquantityonposition-maximum-position parameter?"
            )
        maximum_position: Decimal = args.biasquantityonposition_maximum_position
        target_position: Decimal = (
            args.biasquantityonposition_target_position or Decimal(0)
        )
        return BiasQuantityOnPositionElement(maximum_position, target_position)

    def process_order_pair(
        self,
        context: entropy.Context,
        model_state: ModelState,
        index: int,
        buy: typing.Optional[entropy.Order],
        sell: typing.Optional[entropy.Order],
    ) -> typing.Tuple[typing.Optional[entropy.Order], typing.Optional[entropy.Order]]:
        if buy is not None and sell is not None:
            total_quantity = buy.quantity + sell.quantity
            current_position = model_state.inventory.base.value

            # BUY adjustment formula from the spreadsheet:
            #   =MIN((A2+B2), MAX(0, ((D2-(C2-E2))/D2)*A2))
            biased_buy_quantity = (
                (self.maximum_position - (current_position - self.target_position))
                / self.maximum_position
            ) * buy.quantity
            clamped_biased_buy_quantity = min(
                total_quantity, max(Decimal(0), biased_buy_quantity)
            )
            if buy.quantity != clamped_biased_buy_quantity:
                new_buy = buy.with_update(quantity=clamped_biased_buy_quantity)
                buy_bias_description = (
                    "BUY more"
                    if clamped_biased_buy_quantity > buy.quantity
                    else "BUY less"
                )
                self._logger.debug(
                    f"""BUY order change - maximum position {self.maximum_position} with current position {current_position} and target position {self.target_position} creates a {buy_bias_description} bias:
    Old: {buy}
    New: {new_buy}"""
                )
                buy = new_buy

            # SELL adjustment formula from the spreadsheet:
            #   =MIN((A2+B2), MAX(0, (2-((D2-(C2-E2))/D2))*B2))
            biased_sell_quantity = (
                2
                - (
                    (self.maximum_position - (current_position - self.target_position))
                    / self.maximum_position
                )
            ) * sell.quantity
            clamped_biased_sell_quantity = min(
                total_quantity, max(Decimal(0), biased_sell_quantity)
            )
            if sell.quantity != clamped_biased_sell_quantity:
                new_sell = sell.with_update(quantity=clamped_biased_sell_quantity)
                sell_bias_description = (
                    "SELL more"
                    if clamped_biased_sell_quantity > sell.quantity
                    else "SELL less"
                )
                self._logger.debug(
                    f"""SELL order change - maximum position {self.maximum_position} with current position {current_position} and target position {self.target_position} creates a {sell_bias_description} bias:
    Old: {sell}
    New: {new_sell}"""
                )
                sell = new_sell

        return buy, sell

    def __str__(self) -> str:
        return f"« BiasQuantityOnPositionElement - maximum: {self.maximum_position}, target: {self.target_position} »"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['entropy',
 'entropy.hedging',
 'entropy.layouts',
 'entropy.marketmaking',
 'entropy.marketmaking.orderchain',
 'entropy.oracles',
 'entropy.oracles.ftx',
 'entropy.oracles.market',
 'entropy.oracles.pythnetwork',
 'entropy.oracles.stub',
 'entropy.simplemarketmaking']

package_data = \
{'': ['*']}

install_requires = \
['Rx>=3.2.0,<4.0.0',
 'jsons>=1.6.1,<2.0.0',
 'numpy>=1.22.1,<2.0.0',
 'pandas>=1.4.1,<2.0.0',
 'pyserum>=0.5.0a0,<0.6.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'requests>=2.27.1,<3.0.0',
 'rxpy-backpressure>=1.0.0,<2.0.0',
 'solana>=0.21.0,<0.22.0',
 'websocket-client>=1.2.1,<2.0.0',
 'zstandard>=0.17.0,<0.18.0']

setup_kwargs = {
    'name': 'entropy-explorer',
    'version': '1.0.8',
    'description': 'Python integration for https://entropy.trade',
    'long_description': 'Python API for Entropy Market\n\nGet Started\n\n```\nmake setup\n```\n\nExamples in notebooks\n\n```\ncd notebooks\njupyter notebooks\n```\n\nRun a simple marketmaker\n\n```\ncd entropy\npython simplemarketmaker.py\n```\n\nInstall via pip\n\n```\npip install entropy-explorer\n```\n',
    'author': 'thiccythot',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://entropy.trade',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<3.11',
}


setup(**setup_kwargs)



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import typing

from dataclasses import dataclass
from decimal import Decimal

from .orders import Side


# # 🥭 PlacedOrder tuple
#
# A `PlacedOrder` is a representation of all the data available from an Open Orders account pertaining to a
# particular order.
#
# The information is usually split across 3 collections - 'is bid', 'orders' and 'client ID's. That can be a
# little awkward to use, so this tuple packages it all together, per order.
#
@dataclass
class PlacedOrder:
    id: int
    client_id: int
    side: Side

    @staticmethod
    def build_from_open_orders_data(
        free_slot_bits: Decimal,
        is_bid_bits: Decimal,
        order_ids: typing.Sequence[Decimal],
        client_order_ids: typing.Sequence[Decimal],
    ) -> typing.Sequence["PlacedOrder"]:
        int_free_slot_bits = int(free_slot_bits)
        int_is_bid_bits = int(is_bid_bits)
        placed_orders: typing.List[PlacedOrder] = []
        for index in range(len(order_ids)):
            if not (int_free_slot_bits & (1 << index)):
                order_id = int(order_ids[index])
                client_id = int(client_order_ids[index])
                side = Side.BUY if int_is_bid_bits & (1 << index) else Side.SELL
                placed_orders += [
                    PlacedOrder(id=order_id, client_id=client_id, side=side)
                ]
        return placed_orders

    def __repr__(self) -> str:
        return f"{self}"

    def __str__(self) -> str:
        return f"« PlacedOrder {self.side} [{self.id}] {self.client_id} »"


# # 🥭 PlacedOrdersContainer protocol
#
# The `PlacedOrdersContainer` protocol exposes commonality between the regular Serum `OpenOrders` class and the
# internally-different `PerpOpenOrders` class. Both have their own `placed_orders` member, but are otherwise
# different enough that a common abstract base class would be a bit kludgy.
#
class PlacedOrdersContainer(typing.Protocol):
    placed_orders: typing.Sequence[PlacedOrder]



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import abc
import logging
import requests
import typing

from urllib.parse import unquote

from .output import output


# # 🥭 Notification
#
# This file contains code to send arbitrary notifications.
#

# # 🥭 NotificationTarget class
#
# This base class is the root of the different notification mechanisms.
#
# Derived classes should override `send_notification()` to implement their own sending logic.
#
# Derived classes should not override `send()` since that is the interface outside classes call and it's used to ensure `NotificationTarget`s don't throw an exception when sending.
#
class NotificationTarget(metaclass=abc.ABCMeta):
    def __init__(self) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)

    def send(self, item: typing.Any) -> None:
        try:
            self.send_notification(item)
        except Exception as exception:
            self._logger.error(f"Error sending {item} - {self} - {exception}")

    @abc.abstractmethod
    def send_notification(self, item: typing.Any) -> None:
        raise NotImplementedError(
            "NotificationTarget.send() is not implemented on the base type."
        )

    def __str__(self) -> str:
        return "« NotificationTarget »"

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 TelegramNotificationTarget class
#
# The `TelegramNotificationTarget` sends messages to Telegram.
#
# The format for the telegram notification is:
# 1. The word 'telegram'
# 2. A colon ':'
# 3. The chat ID
# 4. An '@' symbol
# 5. The bot token
#
# For example:
# ```
# telegram:<CHAT-ID>@<BOT-TOKEN>
# ```
#
# The [Telegram instructions to create a bot](https://core.telegram.org/bots#creating-a-new-bot)
# show you how to create the bot token.
class TelegramNotificationTarget(NotificationTarget):
    def __init__(self, address: str) -> None:
        super().__init__()
        chat_id, bot_id = address.split("@", 1)
        self.chat_id = chat_id
        self.bot_id = bot_id

    def send_notification(self, item: typing.Any) -> None:
        payload = {
            "disable_notification": True,
            "chat_id": self.chat_id,
            "text": str(item),
        }
        url = f"https://api.telegram.org/bot{self.bot_id}/sendMessage"
        headers = {"Content-Type": "application/json"}
        requests.post(url, json=payload, headers=headers)

    def __str__(self) -> str:
        return f"« TelegramNotificationTarget Chat ID: {self.chat_id} »"


# # 🥭 DiscordNotificationTarget class
#
# The `DiscordNotificationTarget` sends messages to Discord.
#
class DiscordNotificationTarget(NotificationTarget):
    def __init__(self, address: str) -> None:
        super().__init__()
        self.address = address

    def send_notification(self, item: typing.Any) -> None:
        payload = {"content": str(item)}
        url = self.address
        headers = {"Content-Type": "application/json"}
        requests.post(url, json=payload, headers=headers)

    def __str__(self) -> str:
        return f"« DiscordNotificationTarget Address: {self.address} »"


# # 🥭 MailjetNotificationTarget class
#
# The `MailjetNotificationTarget` sends an email through [Mailjet](https://mailjet.com).
#
# In order to pass everything in to the notifier as a single string (needed to stop
# command-line parameters form getting messy), `MailjetNotificationTarget` requires a
# compound string, separated by colons.
# ```
# mailjet:<MAILJET-API-KEY>:<MAILJET-API-SECRET>:FROM-NAME:FROM-ADDRESS:TO-NAME:TO-ADDRESS
#
# ```
# Individual components are URL-encoded (so, for example, spaces are replaces with %20,
# colons are replaced with %3A).
#
# * `<MAILJET-API-KEY>` and `<MAILJET-API-SECRET>` are from your [Mailjet](https://mailjet.com) account.
# * `FROM-NAME` and `TO-NAME` are just text fields that are used as descriptors in the email messages.
# * `FROM-ADDRESS` is the address the email appears to come from. This must be validated with [Mailjet](https://mailjet.com).
# * `TO-ADDRESS` is the destination address - the email account to which the email is being sent.
#
# Mailjet provides a client library, but really we don't need or want more dependencies. This`
# code just replicates the `curl` way of doing things:
# ```
# curl -s \
# 	-X POST \
# 	--user "$MJ_APIKEY_PUBLIC:$MJ_APIKEY_PRIVATE" \
# 	https://api.mailjet.com/v3.1/send \
# 	-H 'Content-Type: application/json' \
# 	-d '{
#       "SandboxMode":"true",
#       "Messages":[
#         {
#           "From":[
#             {
#               "Email":"pilot@mailjet.com",
#               "Name":"Your Mailjet Pilot"
#             }
#           ],
#           "HTMLPart":"<h3>Dear passenger, welcome to Mailjet!</h3><br />May the delivery force be with you!",
#           "Subject":"Your email flight plan!",
#           "TextPart":"Dear passenger, welcome to Mailjet! May the delivery force be with you!",
#           "To":[
#             {
#               "Email":"passenger@mailjet.com",
#               "Name":"Passenger 1"
#             }
#           ]
#         }
#       ]
# 	}'
# ```
class MailjetNotificationTarget(NotificationTarget):
    def __init__(self, encoded_parameters: str) -> None:
        super().__init__()
        self.address = "https://api.mailjet.com/v3.1/send"
        (
            api_key,
            api_secret,
            subject,
            from_name,
            from_address,
            to_name,
            to_address,
        ) = encoded_parameters.split(":")
        self.api_key: str = unquote(api_key)
        self.api_secret: str = unquote(api_secret)
        self.subject: str = unquote(subject)
        self.from_name: str = unquote(from_name)
        self.from_address: str = unquote(from_address)
        self.to_name: str = unquote(to_name)
        self.to_address: str = unquote(to_address)

    def send_notification(self, item: typing.Any) -> None:
        payload = {
            "Messages": [
                {
                    "From": {"Email": self.from_address, "Name": self.from_name},
                    "Subject": self.subject,
                    "TextPart": str(item),
                    "To": [{"Email": self.to_address, "Name": self.to_name}],
                }
            ]
        }

        url = self.address
        headers = {"Content-Type": "application/json"}
        requests.post(
            url, json=payload, headers=headers, auth=(self.api_key, self.api_secret)
        )

    def __str__(self) -> str:
        return f"« MailjetNotificationTarget To: '{self.to_name}' '{self.to_address}' with subject '{self.subject}' »"


# # 🥭 FilteringNotificationTarget class
#
# This class takes a `NotificationTarget` and a filter function, and only calls the
# `NotificationTarget` if the filter function returns `True` for the notification item.
#
class FilteringNotificationTarget(NotificationTarget):
    def __init__(
        self,
        inner_notifier: NotificationTarget,
        filter_func: typing.Callable[[typing.Any], bool],
    ) -> None:
        super().__init__()
        self.inner_notifier: NotificationTarget = inner_notifier
        self.filter_func = filter_func

    def send_notification(self, item: typing.Any) -> None:
        if self.filter_func(item):
            self.inner_notifier.send_notification(item)

    def __str__(self) -> str:
        return f"« FilteringNotificationTarget For: {self.inner_notifier} »"


# # 🥭 ConsoleNotificationTarget class
#
# The `ConsoleNotificationTarget` prints messages on the console.
#
class ConsoleNotificationTarget(NotificationTarget):
    def __init__(self, name: str) -> None:
        super().__init__()
        self.name = name

    def send_notification(self, item: typing.Any) -> None:
        output(self.name, item)

    def __str__(self) -> str:
        return f"« ConsoleNotificationTarget '{self.name}' »"


# # 🥭 CompoundNotificationTarget class
#
# The `CompoundNotificationTarget` acts as a single `NotificationTarget`, sending notifications to all
# inner `NotificationTarget`s.
#
class CompoundNotificationTarget(NotificationTarget):
    def __init__(self, targets: typing.Sequence[NotificationTarget]) -> None:
        super().__init__()
        self.targets: typing.Sequence[NotificationTarget] = targets
        self.in_exception_handler: bool = False

    def send_notification(self, item: typing.Any) -> None:
        for target in self.targets:
            try:
                target.send(item)
            except Exception as exception:
                if not self.in_exception_handler:
                    self.in_exception_handler = True
                    self._logger.error(
                        f"Failed to send notification to: {target} - {exception}"
                    )
            finally:
                self.in_exception_handler = False

    def __str__(self) -> str:
        inner: typing.List[str] = []
        for target in self.targets:
            inner += [f"{target}"]
        inner_text: str = "\n    ".join(inner)
        return f"""« CompoundNotificationTarget with {len(self.targets)} inner targets:
    {inner_text}
»"""


# # 🥭 parse_notification_target() function
#
# `parse_notification_target()` takes a parameter as a string and returns a notification
# target.
#
# This is most likely used when parsing command-line arguments - this function can be used
# in the `type` parameter of an `add_argument()` call.
#
def parse_notification_target(target: str) -> NotificationTarget:
    protocol, destination = target.split(":", 1)

    if protocol == "telegram":
        return TelegramNotificationTarget(destination)
    elif protocol == "discord":
        return DiscordNotificationTarget(destination)
    elif protocol == "mailjet":
        return MailjetNotificationTarget(destination)
    elif protocol == "console":
        return ConsoleNotificationTarget(destination)
    else:
        raise Exception(f"Unknown protocol: {protocol}")


# # 🥭 NotificationHandler class
#
# A bridge between the worlds of notifications and logging. This allows any
# `NotificationTarget` to be plugged in to the `logging` subsystem to receive log messages
# and notify however it chooses.
#
class NotificationHandler(logging.StreamHandler):  # type: ignore
    def __init__(self, target: NotificationTarget) -> None:
        super().__init__()
        self.target = target

    def emit(self, record: logging.LogRecord) -> None:
        # Don't send error logging from solanaweb3
        if record.name == "solanaweb3.rpc.httprpc.HTTPClient":
            return
        message = self.format(record)
        self.target.send_notification(message)

    def __str__(self) -> str:
        return "« NotificationHandler »"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import logging
import rx
import rx.subject
import types
import typing

from datetime import datetime
from rx.core.typing import Disposable as RxDisposable
from rxpy_backpressure import BackPressure

from .datetimes import local_now
from .output import output


# # 🥭 Observables
#
# This notebook contains some useful shared tools to work with
# [RX Observables](https://rxpy.readthedocs.io/en/latest/reference_observable.html).
#

# # 🥭 NullObserverSubscriber class
#
# This class can subscribe to an `Observable` to do nothing but make sure it runs.
#
class NullObserverSubscriber(rx.core.observer.observer.Observer):
    def __init__(self) -> None:
        super().__init__()

    def on_next(self, item: typing.Any) -> None:
        pass

    def on_error(self, ex: Exception) -> None:
        pass

    def on_completed(self) -> None:
        pass


# # 🥭 PrintingObserverSubscriber class
#
# This class can subscribe to an `Observable` and print out each item.
#
class PrintingObserverSubscriber(rx.core.observer.observer.Observer):
    def __init__(self, report_no_output: bool) -> None:
        super().__init__()
        self.report_no_output = report_no_output
        self.counter = 0

    def on_next(self, item: typing.Any) -> None:
        self.report_no_output = False
        output(self.counter, item)
        self.counter += 1

    def on_error(self, ex: Exception) -> None:
        self.report_no_output = False
        output(ex)

    def on_completed(self) -> None:
        if self.report_no_output:
            output("No items to show.")


# # 🥭 TimestampedPrintingObserverSubscriber class
#
# Just like `PrintingObserverSubscriber` but it puts a timestamp on each printout.
#
class TimestampedPrintingObserverSubscriber(PrintingObserverSubscriber):
    def __init__(self, report_no_output: bool) -> None:
        super().__init__(report_no_output)

    def on_next(self, item: typing.Any) -> None:
        super().on_next(f"{local_now()}: {item}")


# # 🥭 CollectingObserverSubscriber class
#
# This class can subscribe to an `Observable` and collect each item.
#
class CollectingObserverSubscriber(rx.core.observer.observer.Observer):
    def __init__(self) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.collected: typing.List[typing.Any] = []

    def on_next(self, item: typing.Any) -> None:
        self.collected += [item]

    def on_error(self, ex: Exception) -> None:
        self._logger.error(f"Received error: {ex}")

    def on_completed(self) -> None:
        pass


# # 🥭 CaptureFirstItem class
#
# This captures the first item to pass through the pipeline, allowing it to be inspected
# later.
#
class CaptureFirstItem:
    def __init__(self) -> None:
        self.captured: typing.Any = None
        self.has_captured: bool = False

    def capture_if_first(self, item: typing.Any) -> typing.Any:
        if not self.has_captured:
            self.captured = item
            self.has_captured = True

        return item


# # 🥭 TItem type parameter
#
# The `TItem` type parameter is the type parameter for the generic `LatestItemObserverSubscriber`.
#
TItem = typing.TypeVar("TItem")


# # 🥭 LatestItemObserverSubscriber class
#
# This class can subscribe to an `Observable` and capture the latest item as it is observed.
#
class LatestItemObserverSubscriber(
    rx.core.observer.observer.Observer, typing.Generic[TItem]
):
    def __init__(self, initial: TItem) -> None:
        super().__init__()
        self.latest: TItem = initial
        self.update_timestamp: datetime = local_now()

    def on_next(self, item: TItem) -> None:
        self.latest = item
        self.update_timestamp = local_now()

    def on_error(self, ex: Exception) -> None:
        pass

    def on_completed(self) -> None:
        pass


# # 🥭 FunctionObserver
#
# This class takes functions for `on_next()`, `on_error()` and `on_completed()` and returns
# an `Observer` object.
#
# This is mostly for libraries (like `rxpy_backpressure`) that take observers but not their
# component functions.
#
class FunctionObserver(rx.core.observer.observer.Observer):
    def __init__(
        self,
        on_next: typing.Callable[[typing.Any], None],
        on_error: typing.Callable[[Exception], None] = lambda _: None,
        on_completed: typing.Callable[[], None] = lambda: None,
    ) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self._on_next = on_next
        self._on_error = on_error
        self._on_completed = on_completed

    def on_next(self, value: typing.Any) -> None:
        try:
            self._on_next(value)
        except Exception as exception:
            self._logger.warning(f"on_next callable raised exception: {exception}")

    def on_error(self, error: Exception) -> None:
        try:
            self._on_error(error)
        except Exception as exception:
            self._logger.warning(f"on_error callable raised exception: {exception}")

    def on_completed(self) -> None:
        try:
            self._on_completed()
        except Exception as exception:
            self._logger.warning(f"on_completed callable raised exception: {exception}")


# # 🥭 DisposingSubject
#
# This class is a regular Subject that can take additional `RxDisposable` objects to dispose of when the
# `Subject` is being cleaned up.
#
class DisposingSubject(rx.subject.subject.Subject):
    def __init__(self) -> None:
        super().__init__()
        self._to_dispose: typing.List[RxDisposable] = []

    def add_disposable(self, disposable: RxDisposable) -> None:
        self._to_dispose += [disposable]

    def dispose(self) -> None:
        for disposable in self._to_dispose:
            disposable.dispose()
        super().dispose()


# # 🥭 create_backpressure_skipping_observer function
#
# Creates an `Observer` that skips inputs if they are building up while a subscriber works.
#
# This is useful for situations that, say, poll every second but the operation can sometimes
# take multiple seconds to complete. In that case, the latest item will be immediately
# emitted and the in-between items skipped.
#
def create_backpressure_skipping_observer(
    on_next: typing.Callable[[typing.Any], None],
    on_error: typing.Callable[[Exception], None] = lambda _: None,
    on_completed: typing.Callable[[], None] = lambda: None,
) -> rx.core.typing.Observer[typing.Any]:
    observer = FunctionObserver(
        on_next=on_next, on_error=on_error, on_completed=on_completed
    )
    return typing.cast(
        rx.core.typing.Observer[typing.Any], BackPressure.LATEST(observer)
    )


# # 🥭 debug_print_item function
#
# This is a handy item that can be added to a pipeline to show what is being passed at that particular stage. For example, this shows how to print the item before and after filtering:
# ```
# fetch().pipe(
#     ops.map(debug_print_item("Unfiltered:")),
#     ops.filter(lambda item: item.something is not None),
#     ops.map(debug_print_item("Filtered:")),
#     ops.filter(lambda item: item.something_else()),
#     ops.map(act_on_item)
# ).subscribe(some_subscriber)
# ```
#
def debug_print_item(title: str) -> typing.Callable[[typing.Any], typing.Any]:
    def _debug_print_item(item: typing.Any) -> typing.Any:
        output(title, item)
        return item

    return _debug_print_item


# # 🥭 log_subscription_error function
#
# Logs subscription exceptions to the root logger.
#
def log_subscription_error(error: Exception) -> None:
    logging.error(f"Observable subscription error: {error}")


# # 🥭 observable_pipeline_error_reporter function
#
# This intercepts and re-raises an exception, to help report on errors.
#
# RxPy pipelines are tricky to restart, so it's often easier to use the `ops.retry()`
# function in the pipeline. That just swallows the error though, so there's no way to know
# what was raised to cause the retry.
#
# Enter `observable_pipeline_error_reporter()`! Put it in a `catch` just before the `retry`
# and it should log the error properly.
#
# For example:
# ```
# from rx import of, operators as ops
#
# def raise_on_every_third(item):
#     if (item % 3 == 0):
#         raise Exception("Divisible by 3")
#     else:
#         return item
#
# sub1 = of(1, 2, 3, 4, 5, 6).pipe(
#     ops.map(lambda e : raise_on_every_third(e)),
#     ops.catch(observable_pipeline_error_reporter),
#     ops.retry(3)
# )
# sub1.subscribe(lambda item: print(item), on_error = lambda error: print(f"Error : {error}"))
# ```
#
def observable_pipeline_error_reporter(
    ex: Exception, _: rx.core.observable.observable.Observable
) -> rx.core.observable.observable.Observable:
    logging.error(f"Intercepted error in observable pipeline: {ex}")
    raise ex


# # 🥭 TEventDatum type parameter
#
# The `TEventDatum` type parameter is the type parameter for the generic `LatestItemObserverSubscriber`.
#
TEventDatum = typing.TypeVar("TEventDatum")


# # 🥭 EventSource class
#
# A strongly(ish)-typed event source that can handle many subscribers.
#
class EventSource(rx.subject.subject.Subject, typing.Generic[TEventDatum]):
    def __init__(self) -> None:
        super().__init__()
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)

    def on_next(self, event: TEventDatum) -> None:
        super().on_next(event)

    def on_error(self, ex: Exception) -> None:
        super().on_error(ex)

    def on_completed(self) -> None:
        super().on_completed()

    def publish(self, event: TEventDatum) -> None:
        try:
            self.on_next(event)
        except Exception as exception:
            self._logger.warning(f"Failed to publish event '{event}' - {exception}")

    def dispose(self) -> None:
        super().dispose()


# # 🥭 Disposable class
#
# A `RxDisposable` class that can 'fan out' `dispose()` calls to perform additional
# cleanup actions.
#
class Disposable(RxDisposable):
    def __init__(self) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.disposables: typing.List[RxDisposable] = []

    def __enter__(self) -> "Disposable":
        return self

    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ) -> None:
        self.dispose()

    def add_disposable(self, disposable: RxDisposable) -> None:
        self.disposables += [disposable]

    def dispose(self) -> None:
        for index, disposable in enumerate(reversed(self.disposables)):
            try:
                disposable.dispose()
            except Exception as ex:
                self._logger.warning(f"Exception disposing at index {index} - {ex}")


# # 🥭 DisposeWrapper class
#
# A `RxDisposable` class that wraps a lambda to perform some cleanup actions when it is disposed.
#
class DisposeWrapper(RxDisposable):
    def __init__(self, callable: typing.Callable[[], None]) -> None:
        self.callable: typing.Callable[[], None] = callable

    def dispose(self) -> None:
        self.callable()



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import argparse
import entropy
import typing

from decimal import Decimal

from .element import Element
from ...modelstate import ModelState


# # 🥭 PreventPostOnlyCrossingBookElement class
#
# May modifiy an `Order`s price if it would result in too small a difference from the mid-price, meaning less
# of a charge if that `Order` is filled.
#
class PreventPostOnlyCrossingBookElement(Element):
    def __init__(self) -> None:
        super().__init__()

    @staticmethod
    def add_command_line_parameters(parser: argparse.ArgumentParser) -> None:
        pass

    @staticmethod
    def from_command_line_parameters(
        args: argparse.Namespace,
    ) -> "PreventPostOnlyCrossingBookElement":
        return PreventPostOnlyCrossingBookElement()

    def process(
        self,
        context: entropy.Context,
        model_state: ModelState,
        orders: typing.Sequence[entropy.Order],
    ) -> typing.Sequence[entropy.Order]:
        new_orders: typing.List[entropy.Order] = []
        for order in orders:
            if order.order_type == entropy.OrderType.POST_ONLY:
                top_bid: typing.Optional[Decimal] = (
                    model_state.top_bid.price
                    if model_state.top_bid is not None
                    else None
                )
                top_ask: typing.Optional[Decimal] = (
                    model_state.top_ask.price
                    if model_state.top_ask is not None
                    else None
                )
                if (
                    order.side == entropy.Side.BUY
                    and top_ask is not None
                    and order.price >= top_ask
                ):
                    new_buy_price: Decimal = (
                        top_ask - model_state.market.lot_size_converter.tick_size
                    )
                    new_buy: entropy.Order = order.with_update(price=new_buy_price)
                    self._logger.debug(
                        f"""Order change - would cross the orderbook {top_bid} / {top_ask}:
    Old: {order}
    New: {new_buy}"""
                    )
                    new_orders += [new_buy]
                elif (
                    order.side == entropy.Side.SELL
                    and top_bid is not None
                    and order.price <= top_bid
                ):
                    new_sell_price: Decimal = (
                        top_bid + model_state.market.lot_size_converter.tick_size
                    )
                    new_sell: entropy.Order = order.with_update(price=new_sell_price)
                    self._logger.debug(
                        f"""Order change - would cross the orderbook {top_bid} / {top_ask}:
    Old: {order}
    New: {new_sell}"""
                    )

                    new_orders += [new_sell]
                else:
                    # All OK with current order
                    new_orders += [order]
            else:
                # Only change POST_ONLY orders.
                new_orders += [order]

        return new_orders

    def __str__(self) -> str:
        return "« PreventPostOnlyCrossingBookElement »"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import typing

from pyserum._layouts.instructions import InstructionType as PySerumInstructionType
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction

from .instructiontype import InstructionType
from .layouts import layouts
from .entropyinstruction import EntropyInstruction


# # 🥭 InstructionReporter class
#
# The `InstructionReporter` class tries to load and present a decent readable interpretation of a Solana
# instruction.
#
class InstructionReporter:
    def matches(self, instruction: TransactionInstruction) -> bool:
        return True

    def report(self, instruction: TransactionInstruction) -> str:
        instruction_data = "".join("{:02x}".format(x) for x in instruction.data)
        keys: typing.List[str] = []
        for index, key in enumerate(instruction.keys):
            is_writable: str = "Writable " if key.is_writable else "Read-Only"
            is_signer: str = "Signer" if key.is_signer else "      "
            pubkey: str = str(key.pubkey)
            keys += [f"    Key[{index: >2}]: {pubkey: <45} {is_writable} {is_signer}"]
        key_details: str = "\n".join(keys)
        return f"""« Unknown Instruction:
    Program ID: {instruction.program_id}
    Data: {instruction_data}
{key_details}
»"""


# # 🥭 SerumInstructionReporter class
#
# The `SerumInstructionParser` class knows a bit more about Serum instructions.
#
class SerumInstructionReporter(InstructionReporter):
    def __init__(self, serum_program_address: PublicKey) -> None:
        super().__init__()
        self.serum_program_address: PublicKey = serum_program_address

    def matches(self, instruction: TransactionInstruction) -> bool:
        return instruction.program_id == self.serum_program_address

    def report(self, instruction: TransactionInstruction) -> str:
        initial = layouts.SERUM_INSTRUCTION_VARIANT_FINDER.parse(instruction.data)
        instruction_type = PySerumInstructionType(initial.variant)
        instruction_data = "".join("{:02x}".format(x) for x in instruction.data)
        data = "".join("{:02x}".format(x) for x in instruction.data)
        keys: typing.List[str] = []
        for index, key in enumerate(instruction.keys):
            is_writable: str = "Writable " if key.is_writable else "Read-Only"
            is_signer: str = "Signer" if key.is_signer else "      "
            pubkey: str = str(key.pubkey)
            keys += [f"    Key[{index: >2}]: {pubkey: <45} {is_writable} {is_signer}"]
        key_details: str = "\n".join(keys)
        return f"""« Serum Instruction: {instruction_type.name}: {data}
    Program ID: {instruction.program_id}
    Data: {instruction_data}
{key_details}
»"""


# # 🥭 EntropyInstructionReporter class
#
# The `EntropyInstructionReporter` class knows a bit more about Entropy instructions.
#
class EntropyInstructionReporter(InstructionReporter):
    def __init__(self, entropy_program_address: PublicKey) -> None:
        super().__init__()
        self.entropy_program_address: PublicKey = entropy_program_address

    def matches(self, instruction: TransactionInstruction) -> bool:
        return instruction.program_id == self.entropy_program_address

    def report(self, instruction: TransactionInstruction) -> str:
        initial = layouts.MANGO_INSTRUCTION_VARIANT_FINDER.parse(instruction.data)
        parser = layouts.InstructionParsersByVariant[initial.variant]
        if parser is None:
            raise Exception(
                f"Could not find instruction parser for variant {initial.variant} / {InstructionType(initial.variant)}."
            )

        accounts: typing.List[PublicKey] = list(
            map(lambda meta: meta.pubkey, instruction.keys)
        )
        parsed = parser.parse(instruction.data)
        instruction_type = InstructionType(int(parsed.variant))

        entropy_instruction = EntropyInstruction(
            instruction.program_id, instruction_type, instruction.data, parsed, accounts
        )

        return f"{entropy_instruction}"


# # 🥭 CompoundInstructionReporter class
#
# The `CompoundInstructionReporter` class can combine multiple `InstructionReporter`s and pick the right one.
#
class CompoundInstructionReporter(InstructionReporter):
    def __init__(self, reporters: typing.Sequence[InstructionReporter]) -> None:
        super().__init__()
        self.reporters: typing.Sequence[InstructionReporter] = reporters

    def matches(self, instruction: TransactionInstruction) -> bool:
        for reporter in self.reporters:
            if reporter.matches(instruction):
                return True
        return False

    def report(self, instruction: TransactionInstruction) -> str:
        for reporter in self.reporters:
            if reporter.matches(instruction):
                return reporter.report(instruction)
        raise Exception(
            f"Could not find instruction reporter for instruction {instruction}."
        )

    @staticmethod
    def from_addresses(
        entropy_program_address: PublicKey, serum_program_address: PublicKey
    ) -> InstructionReporter:
        base: InstructionReporter = InstructionReporter()
        serum: InstructionReporter = SerumInstructionReporter(serum_program_address)
        entropy: InstructionReporter = EntropyInstructionReporter(
            entropy_program_address
        )
        return CompoundInstructionReporter([entropy, serum, base])



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import typing


# # 🥭 TWatched generic type
#
# Just a generic type name for watchers.
#
TWatched = typing.TypeVar("TWatched", covariant=True)


# # 🥭 Watcher protocol
#
# The `Watcher` protocol just says the object has a property `latest` of type `TWatched`.
#
# The idea behind this is that we can have an object that's subscribed to a websocket and updates, say,
# the Group. But instead of having to explicitly reference websocket implementation details or force
# a common base class on all watchers, we can use the protocol to have any object be a `Watcher` as long
# as it provides a `latest` property of the appropriate type.
#
class Watcher(typing.Protocol[TWatched]):
    @property
    def latest(self) -> TWatched:
        raise NotImplementedError("Watcher.latest is not implemented on the Protocol.")


# # 🥭 ManualUpdateWatcher class
#
# The `ManualUpdateWatcher` class provides a basic implementation of the `Watcher` protocol that
# requires manual intervention (updating of the `value` member) to set the latest version.
#
class ManualUpdateWatcher(typing.Generic[TWatched]):
    def __init__(self, value: TWatched) -> None:
        self.value: TWatched = value

    @property
    def latest(self) -> TWatched:
        return self.value


# # 🥭 LamdaUpdateWatcher class
#
# The `LamdaUpdateWatcher` class provides a basic implementation of the `Watcher` protocol that
# calls a lambda to determine the latest version.
#
class LamdaUpdateWatcher(typing.Generic[TWatched]):
    def __init__(self, accessor: typing.Callable[[], TWatched]) -> None:
        self.accessor: typing.Callable[[], TWatched] = accessor

    @property
    def latest(self) -> TWatched:
        return self.accessor()



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import logging
import multiprocessing
import requests
import types
import typing

from decimal import Decimal
from rx.scheduler.threadpoolscheduler import ThreadPoolScheduler
from solana.publickey import PublicKey
from solana.rpc.commitment import Commitment

from .client import (
    BetterClient,
    ClusterUrlData,
    TransactionMonitor,
    NullTransactionMonitor,
)
from .constants import EntropyConstants
from .idgenerator import IdGenerator, MonotonicIdGenerator
from .instructionreporter import InstructionReporter, CompoundInstructionReporter
from .instrumentlookup import InstrumentLookup
from .marketlookup import MarketLookup
from .text import indent_collection_as_str, indent_item_by
from .tokens import Instrument, Token


# # 🥭 Context class
#
# A `Context` object to manage Solana connection and Entropy configuration.
#
class Context:
    def __init__(
        self,
        name: str,
        cluster_name: str,
        cluster_urls: typing.Sequence[ClusterUrlData],
        skip_preflight: bool,
        tpu_retransmissions: int,
        commitment: str,
        encoding: str,
        blockhash_cache_duration: int,
        http_request_timeout: float,
        stale_data_pauses_before_retry: typing.Sequence[float],
        entropy_program_address: PublicKey,
        serum_program_address: PublicKey,
        group_name: str,
        group_address: PublicKey,
        gma_chunk_size: Decimal,
        gma_chunk_pause: Decimal,
        reflink: typing.Optional[PublicKey],
        instrument_lookup: InstrumentLookup,
        market_lookup: MarketLookup,
        transaction_monitor: TransactionMonitor = NullTransactionMonitor(),
    ) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.name: str = name
        instruction_reporter: InstructionReporter = (
            CompoundInstructionReporter.from_addresses(
                entropy_program_address, serum_program_address
            )
        )
        self.client: BetterClient = BetterClient.from_configuration(
            name,
            cluster_name,
            cluster_urls,
            Commitment(commitment),
            skip_preflight,
            tpu_retransmissions,
            encoding,
            blockhash_cache_duration,
            http_request_timeout,
            stale_data_pauses_before_retry,
            instruction_reporter,
            transaction_monitor,
        )
        self.entropy_program_address: PublicKey = entropy_program_address
        self.serum_program_address: PublicKey = serum_program_address
        self.group_name: str = group_name
        self.group_address: PublicKey = group_address
        self.gma_chunk_size: Decimal = gma_chunk_size
        self.gma_chunk_pause: Decimal = gma_chunk_pause
        self.reflink: typing.Optional[PublicKey] = reflink
        self.instrument_lookup: InstrumentLookup = instrument_lookup
        self.market_lookup: MarketLookup = market_lookup

        self.ping_interval: int = 10

        self.__id_generator: IdGenerator = MonotonicIdGenerator()

        # kangda said in Discord: https://discord.com/channels/791995070613159966/836239696467591186/847816026245693451
        # "I think you are better off doing 4,8,16,20,30"
        self.retry_pauses: typing.Sequence[Decimal] = [
            Decimal(4),
            Decimal(8),
            Decimal(16),
            Decimal(20),
            Decimal(30),
        ]

    def __enter__(self) -> "Context":
        return self

    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ) -> None:
        self.dispose()

    def dispose(self) -> None:
        self.client.dispose()

    def create_thread_pool_scheduler(self) -> ThreadPoolScheduler:
        return ThreadPoolScheduler(multiprocessing.cpu_count())

    def generate_client_id(self) -> int:
        return self.__id_generator.generate_id()

    def lookup_group_name(self, group_address: PublicKey) -> str:
        group_address_str = str(group_address)
        for group in EntropyConstants["groups"]:
            if (
                group["cluster"] == self.client.cluster_name
                and group["publicKey"] == group_address_str
            ):
                return str(group["name"])

        return "« Unknown Group »"

    def instrument(self, symbol: str) -> Instrument:
        instrument = self.instrument_lookup.find_by_symbol(symbol)
        if instrument is None:
            raise Exception(f"Could not find instrument with symbol '{symbol}'.")
        return instrument

    def token(self, symbol: str) -> Token:
        instrument = self.instrument_lookup.find_by_symbol(symbol)
        if instrument is None:
            raise Exception(f"Could not find token with symbol '{symbol}'.")
        return Token.ensure(instrument)

    def fetch_stats(self, url_suffix: str) -> typing.Sequence[typing.Any]:
        stats_url = f"https://stats.entropy.trade/{url_suffix}"
        stats_response = requests.get(stats_url)
        return typing.cast(typing.Sequence[typing.Any], stats_response.json())

    def __str__(self) -> str:
        cluster_urls: str = indent_item_by(
            indent_collection_as_str(self.client.cluster_urls)
        )
        return f"""« Context '{self.name}':
    Cluster Name: {self.client.cluster_name}
    Cluster URLs:
        {cluster_urls}
    Group Name: {self.group_name}
    Group Address: {self.group_address}
    Entropy Program Address: {self.entropy_program_address}
    Serum Program Address: {self.serum_program_address}
»"""

    def __repr__(self) -> str:
        return f"{self}"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://org-docs.friktion.fi/entropy/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/Friktion-Labs)
#   [Email](mailto:hello@blockworks.foundation)


import typing

from dataclasses import dataclass
from datetime import datetime, timedelta
from dateutil import parser
from decimal import Decimal
from solana.publickey import PublicKey

from .account import Account
from .accountinfo import AccountInfo
from .addressableaccount import AddressableAccount
from .combinableinstructions import CombinableInstructions
from .constants import SYSTEM_PROGRAM_ADDRESS
from .context import Context
from .datetimes import utc_now
from .group import Group
from .instructions import (
    build_entropy_redeem_accrued_instructions,
    build_perp_cancel_all_orders_instructions,
    build_perp_cancel_order_instructions,
    build_perp_consume_events_instructions,
    build_perp_place_order_instructions,
)
from .layouts import layouts
from .loadedmarket import LoadedMarket
from .lotsizeconverter import LotSizeConverter, RaisingLotSizeConverter
from .markets import InventorySource, MarketType, Market
from .marketoperations import MarketInstructionBuilder, MarketOperations
from .metadata import Metadata
from .observables import Disposable
from .orders import Order, OrderBook, OrderType, Side
from .perpeventqueue import (
    PerpEvent,
    PerpEventQueue,
    PerpFillEvent,
    UnseenPerpEventChangesTracker,
)
from .perpmarketdetails import PerpMarketDetails
from .publickey import encode_public_key_for_sorting
from .tokens import Instrument, Token
from .tokenbank import TokenBank
from .wallet import Wallet
from .websocketsubscription import (
    IndividualWebSocketSubscriptionManager,
    WebSocketAccountSubscription,
    WebSocketSubscriptionManager,
)
from .version import Version


# # 🥭 FundingRate class
#
# A simple way to package details of a funding rate in a single object.
#
@dataclass
class FundingRate:
    symbol: str
    rate: Decimal
    oracle_price: Decimal
    open_interest: Decimal
    from_: datetime
    to: datetime

    @staticmethod
    def from_stats_data(
        symbol: str,
        lot_size_converter: LotSizeConverter,
        oldest_stats: typing.Dict[str, typing.Any],
        newest_stats: typing.Dict[str, typing.Any],
    ) -> "FundingRate":
        oldest_short_funding = Decimal(oldest_stats["shortFunding"])
        oldest_long_funding = Decimal(oldest_stats["longFunding"])
        oldest_oracle_price = Decimal(oldest_stats["baseOraclePrice"])
        from_timestamp = parser.parse(oldest_stats["time"]).replace(microsecond=0)

        newest_short_funding = Decimal(newest_stats["shortFunding"])
        newest_long_funding = Decimal(newest_stats["longFunding"])
        newest_oracle_price = Decimal(newest_stats["baseOraclePrice"])
        to_timestamp = parser.parse(newest_stats["time"]).replace(microsecond=0)
        raw_open_interest = Decimal(newest_stats["openInterest"])
        open_interest = (
            lot_size_converter.base_size_lots_to_number(raw_open_interest) / 2
        )

        average_oracle_price = (oldest_oracle_price + newest_oracle_price) / 2
        average_oracle_price = newest_oracle_price

        start_funding = (oldest_long_funding + oldest_short_funding) / 2
        end_funding = (newest_long_funding + newest_short_funding) / 2
        funding_difference = end_funding - start_funding

        funding_in_quote_decimals = lot_size_converter.quote.shift_to_decimals(
            funding_difference
        )

        base_price_in_base_lots = average_oracle_price * lot_size_converter.lot_size
        funding_rate = funding_in_quote_decimals / base_price_in_base_lots
        return FundingRate(
            symbol=symbol,
            rate=funding_rate,
            oracle_price=average_oracle_price,
            open_interest=open_interest,
            from_=from_timestamp,
            to=to_timestamp,
        )

    @property
    def extrapolated_apr(self) -> Decimal:
        # From Max (in Discord DM):
        # hpr * 24 * 365 = apr
        # ...
        # hpr * 8760 = apr
        return self.rate * Decimal(8760)

    @property
    def extrapolated_apy(self) -> Decimal:
        # From Max (in Discord DM):
        # (1 + hpr) ^ 8760 = apy + 1
        return ((Decimal(1) + self.rate) ** Decimal(8760)) - Decimal(1)

    def __str__(self) -> str:
        return f"""« FundingRate {self.symbol} {self.rate:,.8%}
    Period: {self.from_} to {self.to}
    Open Interest: {self.open_interest:,.8f}
    Extrapolated APR: {self.extrapolated_apr:,.8%}
    Extrapolated APY: {self.extrapolated_apy:,.8%}
»"""

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 PerpOrderBookSide class
#
# `PerpOrderBookSide` holds orders for one side of a market.
#
class PerpOrderBookSide(AddressableAccount):
    def __init__(
        self,
        account_info: AccountInfo,
        version: Version,
        meta_data: Metadata,
        perp_market_details: PerpMarketDetails,
        bump_index: Decimal,
        free_list_len: Decimal,
        free_list_head: Decimal,
        root_node: Decimal,
        leaf_count: Decimal,
        nodes: typing.Any,
    ) -> None:
        super().__init__(account_info)
        self.version: Version = version

        self.meta_data: Metadata = meta_data
        self.perp_market_details: PerpMarketDetails = perp_market_details
        self.bump_index: Decimal = bump_index
        self.free_list_len: Decimal = free_list_len
        self.free_list_head: Decimal = free_list_head
        self.root_node: Decimal = root_node
        self.leaf_count: Decimal = leaf_count
        self.nodes: typing.Any = nodes

    @staticmethod
    def from_layout(
        layout: typing.Any,
        account_info: AccountInfo,
        version: Version,
        perp_market_details: PerpMarketDetails,
    ) -> "PerpOrderBookSide":
        meta_data = Metadata.from_layout(layout.meta_data)
        bump_index: Decimal = layout.bump_index
        free_list_len: Decimal = layout.free_list_len
        free_list_head: Decimal = layout.free_list_head
        root_node: Decimal = layout.root_node
        leaf_count: Decimal = layout.leaf_count
        nodes: typing.Any = layout.nodes

        return PerpOrderBookSide(
            account_info,
            version,
            meta_data,
            perp_market_details,
            bump_index,
            free_list_len,
            free_list_head,
            root_node,
            leaf_count,
            nodes,
        )

    @staticmethod
    def parse(
        account_info: AccountInfo, perp_market_details: PerpMarketDetails
    ) -> "PerpOrderBookSide":
        data = account_info.data
        if len(data) != layouts.ORDERBOOK_SIDE.sizeof():
            raise Exception(
                f"PerpOrderBookSide data length ({len(data)}) does not match expected size ({layouts.ORDERBOOK_SIDE.sizeof()})"
            )

        layout = layouts.ORDERBOOK_SIDE.parse(data)
        return PerpOrderBookSide.from_layout(
            layout, account_info, Version.V1, perp_market_details
        )

    @staticmethod
    def load(
        context: Context, address: PublicKey, perp_market_details: PerpMarketDetails
    ) -> "PerpOrderBookSide":
        account_info = AccountInfo.load(context, address)
        if account_info is None:
            raise Exception(
                f"PerpOrderBookSide account not found at address '{address}'"
            )
        return PerpOrderBookSide.parse(account_info, perp_market_details)

    def subscribe(
        self,
        context: Context,
        websocketmanager: WebSocketSubscriptionManager,
        callback: typing.Callable[["PerpOrderBookSide"], None],
    ) -> Disposable:
        def __parser(account_info: AccountInfo) -> PerpOrderBookSide:
            return PerpOrderBookSide.parse(account_info, self.perp_market_details)

        subscription = WebSocketAccountSubscription(context, self.address, __parser)
        websocketmanager.add(subscription)
        subscription.publisher.subscribe(on_next=callback)  # type: ignore[call-arg]

        return subscription

    def orders(self) -> typing.Sequence[Order]:
        if self.leaf_count == 0:
            return []

        if self.meta_data.data_type == layouts.DATA_TYPE.Bids:
            order_side = Side.BUY
        else:
            order_side = Side.SELL

        stack = [self.root_node]
        orders: typing.List[Order] = []
        while len(stack) > 0:
            index = int(stack.pop())
            node = self.nodes[index]
            if node.type_name == "leaf":
                timestamp: datetime = node.timestamp
                expiration = Order.NoExpiration
                if node.time_in_force != 0:
                    expiration = timestamp + timedelta(
                        seconds=float(node.time_in_force)
                    )

                price = node.key["price"]
                quantity = node.quantity

                decimals_differential = (
                    self.perp_market_details.base_instrument.decimals
                    - self.perp_market_details.quote_token.token.decimals
                )
                native_to_ui = Decimal(10) ** decimals_differential
                quote_lot_size = self.perp_market_details.quote_lot_size
                base_lot_size = self.perp_market_details.base_lot_size
                actual_price = price * (quote_lot_size / base_lot_size) * native_to_ui

                base_factor = (
                    Decimal(10) ** self.perp_market_details.base_instrument.decimals
                )
                actual_quantity = (
                    quantity * self.perp_market_details.base_lot_size
                ) / base_factor

                orders += [
                    Order(
                        int(node.key["order_id"]),
                        node.client_order_id,
                        node.owner,
                        order_side,
                        actual_price,
                        actual_quantity,
                        OrderType.UNKNOWN,
                        timestamp=timestamp,
                        expiration=expiration,
                    )
                ]
            elif node.type_name == "inner":
                if order_side == Side.BUY:
                    stack = [*stack, node.children[0], node.children[1]]
                else:
                    stack = [*stack, node.children[1], node.children[0]]
        return orders

    def __str__(self) -> str:
        nodes = "\n        ".join(
            [str(node).replace("\n", "\n        ") for node in self.orders()]
        )
        return f"""« PerpOrderBookSide {self.version} [{self.address}]
    {self.meta_data}
    Perp Market: {self.perp_market_details}
    Bump Index: {self.bump_index}
    Free List: {self.free_list_head} (head) {self.free_list_len} (length)
    Root Node: {self.root_node}
    Leaf Count: {self.leaf_count}
        {nodes}
»"""


# # 🥭 PerpMarket class
#
# This class encapsulates our knowledge of a Entropy perps market.
#
class PerpMarket(LoadedMarket):
    def __init__(
        self,
        entropy_program_address: PublicKey,
        address: PublicKey,
        base: Instrument,
        quote: Token,
        underlying_perp_market: PerpMarketDetails,
    ) -> None:
        super().__init__(
            MarketType.PERP,
            entropy_program_address,
            address,
            InventorySource.ACCOUNT,
            base,
            quote,
            RaisingLotSizeConverter(),
        )
        self.underlying_perp_market: PerpMarketDetails = underlying_perp_market
        self.lot_size_converter: LotSizeConverter = LotSizeConverter(
            base,
            underlying_perp_market.base_lot_size,
            quote,
            underlying_perp_market.quote_lot_size,
        )

    @staticmethod
    def isa(market: Market) -> bool:
        return market.type == MarketType.PERP

    @staticmethod
    def ensure(market: Market) -> "PerpMarket":
        if not PerpMarket.isa(market):
            raise Exception(
                f"Market for {market.fully_qualified_symbol} is not a Perp market"
            )
        return typing.cast(PerpMarket, market)

    @property
    def symbol(self) -> str:
        return f"{self.base.symbol}-PERP"

    @property
    def fully_qualified_symbol(self) -> str:
        return f"perp:{self.symbol}"

    @property
    def group(self) -> Group:
        return self.underlying_perp_market.group

    @property
    def bids_address(self) -> PublicKey:
        return self.underlying_perp_market.bids

    @property
    def asks_address(self) -> PublicKey:
        return self.underlying_perp_market.asks

    @property
    def event_queue_address(self) -> PublicKey:
        return self.underlying_perp_market.event_queue

    def parse_account_info_to_orders(
        self, account_info: AccountInfo
    ) -> typing.Sequence[Order]:
        side: PerpOrderBookSide = PerpOrderBookSide.parse(
            account_info, self.underlying_perp_market
        )
        return side.orders()

    def fetch_funding(self, context: Context) -> FundingRate:
        stats = context.fetch_stats(
            f"perp/funding_rate?mangoGroup={self.group.name}&market={self.symbol}"
        )
        newest_stats = stats[0]
        oldest_stats = stats[-1]

        return FundingRate.from_stats_data(
            self.symbol, self.lot_size_converter, oldest_stats, newest_stats
        )

    def unprocessed_events(self, context: Context) -> typing.Sequence[PerpEvent]:
        event_queue: PerpEventQueue = PerpEventQueue.load(
            context, self.event_queue_address, self.lot_size_converter
        )
        return event_queue.unprocessed_events

    def on_fill(
        self, context: Context, handler: typing.Callable[[PerpFillEvent], None]
    ) -> Disposable:
        def _fill_filter(item: PerpEvent) -> None:
            if isinstance(item, PerpFillEvent):
                handler(item)

        return self.on_event(context, _fill_filter)

    def on_event(
        self, context: Context, handler: typing.Callable[[PerpEvent], None]
    ) -> Disposable:
        disposer = Disposable()
        initial: PerpEventQueue = PerpEventQueue.load(
            context, self.event_queue_address, self.lot_size_converter
        )

        manager = IndividualWebSocketSubscriptionManager(context)
        disposer.add_disposable(manager)

        splitter: UnseenPerpEventChangesTracker = UnseenPerpEventChangesTracker(initial)

        def __split_handler(pev: PerpEventQueue) -> None:
            for event in splitter.unseen(pev):
                handler(event)

        subscription = initial.subscribe(context, manager, __split_handler)
        disposer.add_disposable(subscription)

        manager.open()

        return disposer

    def __str__(self) -> str:
        underlying: str = f"{self.underlying_perp_market}".replace("\n", "\n    ")
        return f"""« PerpMarket {self.symbol} {self.address} [{self.program_address}]
    {underlying}
»"""


# # 🥭 PerpMarketInstructionBuilder
#
# This file deals with building instructions for Perp markets.
#
# As a matter of policy for all InstructionBuidlers, construction and build_* methods should all work with
# existing data, requiring no fetches from Solana or other sources. All necessary data should all be loaded
# on initial setup in the `load()` method.
#
class PerpMarketInstructionBuilder(MarketInstructionBuilder):
    def __init__(
        self,
        context: Context,
        wallet: Wallet,
        perp_market: PerpMarket,
        group: Group,
        account: Account,
    ) -> None:
        super().__init__()
        self.context: Context = context
        self.wallet: Wallet = wallet
        self.perp_market: PerpMarket = perp_market
        self.group: Group = group
        self.account: Account = account
        self.mngo_token_bank: TokenBank = self.group.liquidity_incentive_token_bank

    @staticmethod
    def load(
        context: Context,
        wallet: Wallet,
        perp_market: PerpMarket,
        group: Group,
        account: Account,
    ) -> "PerpMarketInstructionBuilder":
        return PerpMarketInstructionBuilder(
            context, wallet, perp_market, group, account
        )

    def build_cancel_order_instructions(
        self, order: Order, ok_if_missing: bool = False
    ) -> CombinableInstructions:
        return build_perp_cancel_order_instructions(
            self.context,
            self.wallet,
            self.account,
            self.perp_market.underlying_perp_market,
            order,
            ok_if_missing,
        )

    def build_place_order_instructions(self, order: Order) -> CombinableInstructions:
        return build_perp_place_order_instructions(
            self.context,
            self.wallet,
            self.perp_market.underlying_perp_market.group,
            self.account,
            self.perp_market.underlying_perp_market,
            order.price,
            order.quantity,
            order.client_id,
            order.side,
            order.order_type,
            order.reduce_only,
            expiration=order.expiration,
            match_limit=order.match_limit,
            reflink=self.context.reflink,
        )

    def build_settle_instructions(self) -> CombinableInstructions:
        return CombinableInstructions.empty()

    def build_crank_instructions(
        self, addresses: typing.Sequence[PublicKey], limit: Decimal = Decimal(32)
    ) -> CombinableInstructions:
        distinct_addresses: typing.List[PublicKey] = [self.account.address]
        for address in addresses:
            if address not in distinct_addresses:
                distinct_addresses += [address]

        if len(distinct_addresses) > limit:
            self._logger.warn(
                f"Cranking limited to {limit} of {len(distinct_addresses)} addresses waiting to be cranked."
            )

        limited_addresses = distinct_addresses[
            0 : min(int(limit), len(distinct_addresses))
        ]
        limited_addresses.sort(key=encode_public_key_for_sorting)
        self._logger.debug(
            f"About to crank {len(limited_addresses)} addresses: {limited_addresses}"
        )

        return build_perp_consume_events_instructions(
            self.context,
            self.group,
            self.perp_market.underlying_perp_market,
            limited_addresses,
            limit,
        )

    def build_redeem_instructions(self) -> CombinableInstructions:
        return build_entropy_redeem_accrued_instructions(
            self.context,
            self.wallet,
            self.perp_market,
            self.group,
            self.account,
            self.mngo_token_bank,
        )

    def build_cancel_all_orders_instructions(
        self, limit: Decimal = Decimal(32)
    ) -> CombinableInstructions:
        return build_perp_cancel_all_orders_instructions(
            self.context,
            self.wallet,
            self.account,
            self.perp_market.underlying_perp_market,
            limit,
        )

    def __str__(self) -> str:
        return """« PerpMarketInstructionBuilder »"""


# # 🥭 PerpMarketOperations
#
# This file deals with placing orders for Perps.
#
class PerpMarketOperations(MarketOperations):
    def __init__(
        self,
        context: Context,
        wallet: Wallet,
        account: Account,
        market_instruction_builder: PerpMarketInstructionBuilder,
    ) -> None:
        super().__init__(market_instruction_builder.perp_market)
        self.context: Context = context
        self.wallet: Wallet = wallet
        self.market_instruction_builder: PerpMarketInstructionBuilder = (
            market_instruction_builder
        )
        self.account: Account = account

    @staticmethod
    def ensure(market_ops: MarketOperations) -> "PerpMarketOperations":
        if not isinstance(market_ops, PerpMarketOperations):
            raise Exception(
                f"MarketOperations for {market_ops.symbol} is not a PerpMarketOperations"
            )
        return market_ops

    @property
    def perp_market(self) -> PerpMarket:
        return self.market_instruction_builder.perp_market

    def cancel_order(
        self, order: Order, ok_if_missing: bool = False
    ) -> typing.Sequence[str]:
        self._logger.info(f"Cancelling {self.symbol} order {order}.")
        signers: CombinableInstructions = CombinableInstructions.from_wallet(
            self.wallet
        )
        cancel: CombinableInstructions = (
            self.market_instruction_builder.build_cancel_order_instructions(
                order, ok_if_missing=ok_if_missing
            )
        )
        crank = self._build_crank(add_self=True)
        settle = self.market_instruction_builder.build_settle_instructions()
        return (signers + cancel + crank + settle).execute(self.context)

    def place_order(
        self, order: Order, crank_limit: Decimal = Decimal(5)
    ) -> typing.Sequence[str]:
        client_id: int = order.client_id or self.context.generate_client_id()
        signers: CombinableInstructions = CombinableInstructions.from_wallet(
            self.wallet
        )
        order_with_client_id: Order = order.with_update(client_id=client_id)
        self._logger.info(f"Placing {self.symbol} order {order_with_client_id}.")
        place: CombinableInstructions = (
            self.market_instruction_builder.build_place_order_instructions(
                order_with_client_id
            )
        )
        crank = self._build_crank(add_self=True, limit=crank_limit)
        settle = self.market_instruction_builder.build_settle_instructions()
        return (signers + place + crank + settle).execute(self.context)

    def settle(self) -> typing.Sequence[str]:
        signers: CombinableInstructions = CombinableInstructions.from_wallet(
            self.wallet
        )
        settle = self.market_instruction_builder.build_settle_instructions()
        return (signers + settle).execute(self.context)

    def crank(self, limit: Decimal = Decimal(32)) -> typing.Sequence[str]:
        signers: CombinableInstructions = CombinableInstructions.from_wallet(
            self.wallet
        )
        crank = self._build_crank(limit=limit)
        return (signers + crank).execute(self.context)

    def create_openorders(self) -> PublicKey:
        return SYSTEM_PROGRAM_ADDRESS

    def ensure_openorders(self) -> PublicKey:
        return SYSTEM_PROGRAM_ADDRESS

    def load_orderbook(self) -> OrderBook:
        return self.perp_market.fetch_orderbook(self.context)

    def load_my_orders(
        self, cutoff: typing.Optional[datetime] = utc_now()
    ) -> typing.Sequence[Order]:
        orderbook: OrderBook = self.load_orderbook()
        return orderbook.all_orders_for_owner(self.account.address, cutoff)

    def _build_crank(
        self, limit: Decimal = Decimal(32), add_self: bool = False
    ) -> CombinableInstructions:
        accounts_to_crank: typing.List[PublicKey] = []
        for event_to_crank in self.perp_market.unprocessed_events(self.context):
            accounts_to_crank += event_to_crank.accounts_to_crank

        if add_self:
            accounts_to_crank += [self.account.address]

        if len(accounts_to_crank) == 0:
            return CombinableInstructions.empty()

        self._logger.debug(
            f"Building crank instruction with {len(accounts_to_crank)} public keys, throttled to {limit}"
        )
        return self.market_instruction_builder.build_crank_instructions(
            accounts_to_crank, limit
        )

    def __str__(self) -> str:
        return f"""« PerpMarketOperations [{self.symbol}] »"""


# # 🥭 PerpMarketStub class
#
# This class holds information to load a `PerpMarket` object but doesn't automatically load it.
#
class PerpMarketStub(Market):
    def __init__(
        self,
        entropy_program_address: PublicKey,
        address: PublicKey,
        base: Instrument,
        quote: Token,
        group_address: PublicKey,
    ) -> None:
        super().__init__(
            MarketType.STUB,
            entropy_program_address,
            address,
            InventorySource.ACCOUNT,
            base,
            quote,
            RaisingLotSizeConverter(),
        )
        self.group_address: PublicKey = group_address

    @property
    def fully_qualified_symbol(self) -> str:
        return f"perp:{self.symbol}"

    def load(
        self, context: Context, group: typing.Optional[Group] = None
    ) -> PerpMarket:
        actual_group: Group = group or Group.load(context, self.group_address)
        underlying_perp_market: PerpMarketDetails = PerpMarketDetails.load(
            context, self.address, actual_group
        )
        return PerpMarket(
            self.program_address,
            self.address,
            self.base,
            self.quote,
            underlying_perp_market,
        )

    @property
    def symbol(self) -> str:
        return f"{self.base.symbol}-PERP"

    def __str__(self) -> str:
        return (
            f"« PerpMarketStub {self.symbol} {self.address} [{self.program_address}] »"
        )



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import argparse
import typing

from .afteraccumulateddepthelement import AfterAccumulatedDepthElement
from .biasquantityonpositionelement import BiasQuantityOnPositionElement
from .biasquoteelement import BiasQuoteElement
from .biasquoteonpositionelement import BiasQuoteOnPositionElement
from .chain import Chain
from .confidenceintervalelement import ConfidenceIntervalElement
from .element import Element
from .fixedpositionsizeelement import FixedPositionSizeElement
from .fixedspreadelement import FixedSpreadElement
from .maximumquantityelement import MaximumQuantityElement
from .minimumquantityelement import MinimumQuantityElement
from .minimumchargeelement import MinimumChargeElement
from .preventpostonlycrossingbookelement import PreventPostOnlyCrossingBookElement
from .quotesinglesideelement import QuoteSingleSideElement
from .roundtolotsizeelement import RoundToLotSizeElement
from .ratioselement import RatiosElement
from .topofbookelement import TopOfBookElement

_DEFAULT_CHAIN = [
    "confidenceinterval",
    "minimumcharge",
    "biasquoteonposition",
    "preventpostonlycrossingbook",
    "roundtolotsize",
]


# # 🥭 ChainBuilder class
#
# A `ChainBuilder` class to allow building a `Chain`, keeping parameter and constructor complexities all in
# one place.
#
class ChainBuilder:
    @staticmethod
    def add_command_line_parameters(parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--chain",
            type=str,
            action="append",
            default=[],
            help="The specific order chain elements to use instead of the default chain",
        )
        # OrderType is used by multiple elements so specify it here rather than have them fighting over which
        # one specifies it.
        # Now add args for all the elements.
        AfterAccumulatedDepthElement.add_command_line_parameters(parser)
        BiasQuantityOnPositionElement.add_command_line_parameters(parser)
        BiasQuoteElement.add_command_line_parameters(parser)
        BiasQuoteOnPositionElement.add_command_line_parameters(parser)
        ConfidenceIntervalElement.add_command_line_parameters(parser)
        FixedSpreadElement.add_command_line_parameters(parser)
        FixedPositionSizeElement.add_command_line_parameters(parser)
        MaximumQuantityElement.add_command_line_parameters(parser)
        MinimumQuantityElement.add_command_line_parameters(parser)
        MinimumChargeElement.add_command_line_parameters(parser)
        PreventPostOnlyCrossingBookElement.add_command_line_parameters(parser)
        QuoteSingleSideElement.add_command_line_parameters(parser)
        RatiosElement.add_command_line_parameters(parser)
        RoundToLotSizeElement.add_command_line_parameters(parser)
        TopOfBookElement.add_command_line_parameters(parser)

    # This function is the converse of `add_command_line_parameters()` - it takes
    # an argument of parsed command-line parameters and expects to see the ones it added
    # to that collection in the `add_command_line_parameters()` call.
    #
    # It then uses those parameters to create a properly-configured `Chain` object.
    #
    @staticmethod
    def from_command_line_parameters(args: argparse.Namespace) -> Chain:
        chain_names: typing.Sequence[str] = args.chain
        if chain_names is None or len(chain_names) == 0:
            chain_names = _DEFAULT_CHAIN

        elements: typing.List[Element] = []
        for name in chain_names:
            element = ChainBuilder._create_element_by_name(args, name)
            elements += [element]

        return Chain(elements)

    @staticmethod
    def _create_element_by_name(args: argparse.Namespace, name: str) -> Element:
        proper_name: str = name.upper()
        if proper_name == "AFTERACCUMULATEDDEPTH":
            return AfterAccumulatedDepthElement.from_command_line_parameters(args)
        elif proper_name == "BIASQUANTITYONPOSITION":
            return BiasQuantityOnPositionElement.from_command_line_parameters(args)
        elif proper_name == "BIASQUOTE":
            return BiasQuoteElement.from_command_line_parameters(args)
        elif proper_name == "BIASQUOTEONPOSITION":
            return BiasQuoteOnPositionElement.from_command_line_parameters(args)
        elif proper_name == "CONFIDENCEINTERVAL":
            return ConfidenceIntervalElement.from_command_line_parameters(args)
        elif proper_name == "FIXEDSPREAD":
            return FixedSpreadElement.from_command_line_parameters(args)
        elif proper_name == "FIXEDPOSITIONSIZE":
            return FixedPositionSizeElement.from_command_line_parameters(args)
        elif proper_name == "MAXIMUMQUANTITY":
            return MaximumQuantityElement.from_command_line_parameters(args)
        elif proper_name == "MINIMUMCHARGE":
            return MinimumChargeElement.from_command_line_parameters(args)
        elif proper_name == "MINIMUMQUANTITY":
            return MinimumQuantityElement.from_command_line_parameters(args)
        elif proper_name == "PREVENTPOSTONLYCROSSINGBOOK":
            return PreventPostOnlyCrossingBookElement.from_command_line_parameters(args)
        elif proper_name == "QUOTESINGLESIDE":
            return QuoteSingleSideElement.from_command_line_parameters(args)
        elif proper_name == "RATIOS":
            return RatiosElement.from_command_line_parameters(args)
        elif proper_name == "ROUNDTOLOTSIZE":
            return RoundToLotSizeElement.from_command_line_parameters(args)
        elif proper_name == "TOPOFBOOK":
            return TopOfBookElement.from_command_line_parameters(args)
        else:
            raise Exception(f"Unknown chain element: '{proper_name}'")



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import abc
import entropy
import typing

from .element import Element
from ...modelstate import ModelState


# # 🥭 PairwiseElement class
#
# Some `Element`s want to process `Order`s in pairs or levels. Typically they'll have a BUY and a
# SELL at one spread, another BUY and SELL at a second spread, another BUY and SELL at a third
# spread, and so on.
#
# The 'desired orders' that come in don't fit that pattern though. It's just a list of `Order`s.
# The `Order`s may have a pair-wise structure, but it's not imposed upon them.
#
# If the `Order`s _do_ have such a pair-wise structure, some `Element`s can take advantage of that.
#
# (The assumption is that if your `Order`s _don't_ have a pair-wise structure - a matching BUY for
# each SELL - then you don't expect to successfully use `Element`s that depend upon such a
# structure.)
#
# The `PairwiseElement` handles converting from an unstructured list of `Order`s into a pair-wise
# structure, and then calls the derived class's `process_order_pair()` method to process a pair.
#
class PairwiseElement(Element, metaclass=abc.ABCMeta):
    def __init__(self) -> None:
        super().__init__()

    @abc.abstractmethod
    def process_order_pair(
        self,
        context: entropy.Context,
        model_state: ModelState,
        index: int,
        buy: typing.Optional[entropy.Order],
        sell: typing.Optional[entropy.Order],
    ) -> typing.Tuple[typing.Optional[entropy.Order], typing.Optional[entropy.Order]]:
        raise NotImplementedError(
            "PairwiseElement.process_order_pair() is not implemented on the base type."
        )

    # If multiple levels are specified, we want to process them in order.
    #
    # But 'in order' is complicated. The way it will be expected to work is:
    # * First BUY and first SELL are treated as a pair
    # * Second BUY and second SELL are treated as a pair
    # * Third BUY and third SELL are treated as a pair
    # * etc.
    # But (another but!) 'first' means closest to the top of the book to people, not necessarily
    # first in the incoming order list.
    #
    # We want to meet that expected approach, so we'll:
    # * Split the list into BUYs and SELLs
    # * Sort the two lists so closest to top-of-book is at index 0
    # * Call process_order_pair() for each paired BUY and SELL, with the index parameter being
    #   the index into the BUY and SELL lists.
    def process(
        self,
        context: entropy.Context,
        model_state: ModelState,
        orders: typing.Sequence[entropy.Order],
    ) -> typing.Sequence[entropy.Order]:
        buys: typing.List[entropy.Order] = list(
            [order for order in orders if order.side == entropy.Side.BUY]
        )
        buys.sort(key=lambda order: order.price, reverse=True)
        sells: typing.List[entropy.Order] = list(
            [order for order in orders if order.side == entropy.Side.SELL]
        )
        sells.sort(key=lambda order: order.price)

        pair_count: int = max(len(buys), len(sells))
        new_orders: typing.List[entropy.Order] = []
        for index in range(pair_count):
            old_buy: typing.Optional[entropy.Order] = (
                buys[index] if index < len(buys) else None
            )
            old_sell: typing.Optional[entropy.Order] = (
                sells[index] if index < len(sells) else None
            )

            (new_buy, new_sell) = self.process_order_pair(
                context, model_state, index, old_buy, old_sell
            )
            if new_buy is not None:
                new_orders += [new_buy]

            if new_sell is not None:
                new_orders += [new_sell]

        return new_orders

    def __str__(self) -> str:
        return "« PairwiseElement »"
